/*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module holds the C implementation of a simple resource fork sanity checker.
	The code is taken from the Disinfectant source code and was contributed by John Norstad.

	I added pragmas so that I could build with full warnings and yet ignore some of
	the warnings that the MPW C compiler generates for this code.  I really didn't want
	to mess with the code.  Remember the engineer's maxim -- if it works, don't fuck with it.
*/

#include <Types.h>
#include <Files.h>
#include <Errors.h>
#include <Memory.h>

pascal OSErr CheckResourceForkSanity (const FSSpec *fss, Boolean *sane);

pascal OSErr CheckResourceForkSanity (const FSSpec *fss, Boolean *sane)
{
	HParamBlockRec		fBlock;			/* file param block */
	short					refNum;			/* file refnum */
	long					count;			/* number of bytes to read */
	long					logEOF;			/* logical EOF */
	unsigned char		*map;				/* pointer to resource map */
	unsigned long		dataLWA;			/* offset in file of data end */
	unsigned long		mapLWA;			/* offset in file of map end */
	unsigned short		typeFWA;			/* offset from map begin to type list */
	unsigned short		nameFWA;			/* offset from map begin to name list */
	unsigned char		*pType;			/* pointer into type list */
	unsigned char		*pName;			/* pointer to start of name list */
	unsigned char		*pMapEnd;		/* pointer to end of map */
	short					nType;			/* number of resource types in map */
	unsigned char		*pTypeEnd;		/* pointer to end of type list */
	short					nRes;				/* number of resources of given type */
	unsigned short		refFWA;			/* offset from type list to ref list */
	unsigned char		*pRef;			/* pointer into reference list */
	unsigned char		*pRefEnd;		/* pointer to end of reference list */
	unsigned short		resNameFWA;		/* offset from name list to resource name */
	unsigned char		*pResName;		/* pointer to resource name */
	unsigned long		resDataFWA;		/* offset from data begin to resource data */
	Boolean				mapOK;			/* true if map is sane */
	OSErr					rCode;			/* error code */
	
	struct {
		unsigned long		dataFWA;		/* offset in file of data */
		unsigned long		mapFWA;		/* offset in file of map */
		unsigned long		dataLen;		/* data area length */
		unsigned long		mapLen;		/* map area length */
	} header;
	
	/* Open the resource file. */
	
	fBlock.ioParam.ioNamePtr = (unsigned char *) fss->name;
	fBlock.fileParam.ioDirID = fss->parID;
	fBlock.ioParam.ioVRefNum = fss->vRefNum;
	fBlock.ioParam.ioVersNum = 0;
	fBlock.ioParam.ioPermssn = fsRdPerm;
	fBlock.ioParam.ioMisc = nil;
	rCode = PBHOpenRF(&fBlock, false);
	if (rCode) {
		if (rCode == fnfErr) {
			*sane = true;
			return noErr;
		} else {
			return rCode;
		};
	};
	
	/* Get the logical eof of the file. */
	
	refNum = fBlock.ioParam.ioRefNum;
	rCode = GetEOF(refNum, &logEOF);
	if (rCode) return rCode;
	if (!logEOF) {
		*sane = true;
		rCode = FSClose(refNum);
		if (rCode) return rCode;
		return noErr;
	};
	
	/* Read and validate the resource header. */
	
	count = 16;
	rCode = FSRead(refNum, &count, (Ptr)&header);
	if (rCode) {
		FSClose(refNum);
		return rCode;
	};
	dataLWA = header.dataFWA + header.dataLen;
	mapLWA = header.mapFWA + header.mapLen;
	mapOK = count == 16 && header.mapLen > 28 &&
		header.dataFWA < 0x01000000 && header.mapFWA < 0x01000000 &&
		dataLWA <= logEOF && mapLWA <= logEOF &&
		(dataLWA <= header.mapFWA || mapLWA <= header.dataFWA);
		
	/* Read the resource map. */
	
	map = nil;
	if (mapOK) {
		map = (unsigned char *) NewPtr(header.mapLen);
		if (!(rCode = SetFPos(refNum, fsFromStart, header.mapFWA))) {
			count = header.mapLen;
			rCode = FSRead(refNum, &count, map);
		};
	};
	
	/* Verify the type list and name list offsets. */
	
	if (!rCode) {
		typeFWA = *(unsigned short*)(map+24);
		nameFWA = *(unsigned short*)(map+26);
		mapOK = typeFWA == 28 && nameFWA >= typeFWA && nameFWA <= header.mapLen &&
			!(typeFWA & 1) && !(nameFWA & 1);
	};
	
	/* Verify the type list, reference lists, and name list. */
	
	if (mapOK) {
		pType = map + typeFWA;
		pName = map + nameFWA;
		pMapEnd = map + header.mapLen;
#pragma warnings off
		nType = *(short*)pType + 1;
#pragma warnings full
		pType += 2;
		pTypeEnd = pType + (nType<<3);
#pragma warnings off
		if (mapOK = pTypeEnd <= pMapEnd) {
#pragma warnings full
			while (pType < pTypeEnd) {
#pragma warnings off
				nRes = *(short*)(pType+4) + 1;
#pragma warnings full
				refFWA = *(unsigned short*)(pType+6);
				pRef = map + typeFWA + refFWA;
				pRefEnd = pRef + 12*nRes;
				if (!(mapOK = pRef >= pTypeEnd && pRef < pName && 
					!(refFWA & 1))) break;
				while (pRef < pRefEnd) {
					resNameFWA = *(unsigned short*)(pRef+2);
					if (resNameFWA != 0xFFFF) {
						pResName = pName + resNameFWA;
						if (!(mapOK = pResName + *pResName < pMapEnd)) break;
					};
					resDataFWA = *(unsigned long*)(pRef+4) & 0x00FFFFFF;
					if (!(mapOK = header.dataFWA + resDataFWA < dataLWA)) break;
					pRef += 12;
				};
				if (!mapOK) break;
				pType += 8;
			};
		};
	};
	
	/* Dispose of the resource map, close the file and return. */
	
	if (map) DisposePtr( (Ptr) map);
	if (!rCode) {
		rCode = FSClose(refNum);
	} else {
		(void) FSClose(refNum);
	};
	*sane = mapOK;
	return rCode;
}
